/*-
 * %%----------------------------------------------------------------------------------------------
 * Archive ouverte UNIGE - AoU Common - AoUControllerAdvice.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2022 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.aou.controlleradvice;

import static ch.unige.solidify.controller.SolidifyControllerAdvice.StackTraceDisplay.WITHOUT_STACK_TRACE;
import static ch.unige.solidify.controller.SolidifyControllerAdvice.StackTraceDisplay.WITH_STACK_TRACE;

import org.apache.catalina.connector.ClientAbortException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.web.servlet.error.ErrorAttributes;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;

import jakarta.servlet.http.HttpServletRequest;

import ch.unige.solidify.config.SolidifyProperties;
import ch.unige.solidify.controller.SolidifyControllerAdvice;
import ch.unige.solidify.exception.SolidifyError;
import ch.unige.solidify.service.MessageService;

import ch.unige.aou.exception.AoUMetadataParseException;
import ch.unige.aou.exception.AouError;
import ch.unige.aou.exception.AouMetadataPotentialDuplicateException;
import ch.unige.aou.exception.AouMetadataValidationException;
import ch.unige.aou.exception.AouRuntimeException;

@ControllerAdvice
public class AoUControllerAdvice extends SolidifyControllerAdvice {
  private static final Logger log = LoggerFactory.getLogger(AoUControllerAdvice.class);

  private final ErrorAttributes errorAttributes;
  private final HttpHeaders headers = new HttpHeaders();

  public AoUControllerAdvice(ErrorAttributes errorAttributes, MessageService messageService, SolidifyProperties solidifyProperties) {
    super(errorAttributes, messageService, solidifyProperties);
    this.errorAttributes = errorAttributes;
    this.headers.add(HttpHeaders.CONTENT_TYPE, "application/hal+json");
  }

  // ClientAbortException belongs to Tomcat, handle it at project level since Solidify-Backend is container agnostic
  @ExceptionHandler(ClientAbortException.class)
  public void handleClientAbortException(ClientAbortException exception, HttpServletRequest request) {
    final String message = "ClientAbortException generated by request {} {} from remote address {} with X-FORWARDED-FOR {}";
    final String headerXFF = request.getHeader("X-FORWARDED-FOR");
    log.warn(message, request.getMethod(), request.getRequestURL(), request.getRemoteAddr(), headerXFF);
  }

  @ExceptionHandler(AoUMetadataParseException.class)
  public ResponseEntity<SolidifyError> handleException(AoUMetadataParseException exception, HttpServletRequest request) {
    String message = "";
    if (exception.getCause() != null) {
      message = exception.getCause().getMessage();
    }
    return this.buildResponse(exception, request, message, HttpStatus.BAD_REQUEST, WITH_STACK_TRACE);
  }

  @ExceptionHandler(AouMetadataValidationException.class)
  public ResponseEntity<SolidifyError> handleException(AouMetadataValidationException exception, HttpServletRequest request) {

    final SolidifyError solidifyError = new SolidifyError(this.errorAttributes, request);
    solidifyError.setValidationError(exception.getValidationError());
    solidifyError.setMessage(exception.getMessage());
    solidifyError.setStatus(HttpStatus.BAD_REQUEST);

    this.logException(exception, null, request, StackTraceDisplay.WITHOUT_STACK_TRACE);
    return ResponseEntity.badRequest().headers(this.headers).body(solidifyError);
  }

  @ExceptionHandler(AouMetadataPotentialDuplicateException.class)
  public ResponseEntity<AouError> handleException(AouMetadataPotentialDuplicateException exception, HttpServletRequest request) {
    return this.handleAouRuntimeException(exception, request, WITHOUT_STACK_TRACE);
  }

  @ExceptionHandler(AouRuntimeException.class)
  public ResponseEntity<AouError> handleException(AouRuntimeException exception, HttpServletRequest request) {
    return this.handleAouRuntimeException(exception, request, WITH_STACK_TRACE);
  }

  private ResponseEntity<AouError> handleAouRuntimeException(AouRuntimeException exception, HttpServletRequest request,
          StackTraceDisplay withStackTrace) {
    AouError aouError = exception.getAouError();
    aouError.setPath(request.getRequestURI());

    this.logException(exception, null, request, withStackTrace);

    aouError.setMessage(exception.getMessage());
    aouError.setError(exception.getMessage());
    aouError.setStatus(HttpStatus.BAD_REQUEST);

    return (ResponseEntity.status(HttpStatus.BAD_REQUEST).headers(this.headers)).body(aouError);
  }
}
